# -*- coding:utf-8 -*-
'''
title      : image_proc.py

CONTENTS: translation of /NaoDataConverter/include/ImageProc.h 
          and /NaoDataConverter/lib/libImageProc/ImageProc.cpp 

CHANGES : In Python, /NaoDataConverter/include/Blob*.h cannot be use.
          Therefore, cv.FindContours() etc. are used in place of them
          and BLOB* are deleted in this script. I use the moments
          position of each color region in place of the position of 
          region range.

------------------------------ REFERENCES ------------------------------
 [1] ``OpenCV 2.4.8.0 documentation", cited by April 29, 2014,
     available from http://docs.opencv.org/index.html
 [2] ``cplusplus.com - The C++ Resources Network", cited by April 24, 2014,
     available from http://www.cplusplus.com/
 [3] ``8.3. collections - High-performance container datatypes - Python v2.7.6 documentation",
     cited by April 23, 2014, available from http://docs.python.org/2.7/library/collections.html
 [4] ``7.2. re - Regular expression operations - Python v2.7.6 documentation",
     cited by April 24, 2014, available from http://docs.python.org/2.7/library/re.html
 [5] ``cv.CreateHist - Nullege Python Samples", cited by April 24, 2014,
     available from http://nullege.com/codes/search/cv.CreateHist
-------------------------------------------------------------------------

--------------------------- PRIVATE VARIABLES ---------------------------
 ORIGINAL                             CHANGES
 int verbose_                      -> _verbose
 COLOR_CONFIG col_conf_[COLOR_NUM] -> _col_conf
 int intensity_th_                 -> _intensity_th
 int area_th_                      -> _area_th
 IplImage* debug_img_              -> _debug_img
 None                              -> _contours       (NEW VALIABLE)
 None                              -> _contours_range (NEW VALIABLE)
-------------------------------------------------------------------------
'''
import cv
import collections
import sys
import re
import numpy as np

class ImageProc:
    # -------------------- PUBLIC STRUCT AND VARIABLES --------------------

    # Original: enum Colors
    # use dictionary type in place of enum
    color_name = ['RED', 'YELLOW', 'GREEN', 'BLUE', 'COLOR_NUM']
    Colors = {}
    for i in xrange(0, len(color_name)):
        Colors.setdefault(color_name[i], i)
    del i

    # Original: struct color_config
    # int range_h[2] -> range_h
    # int range_s[2] -> range_s
    # int range_v[2] -> range_v
    COLOR_CONFIG = collections.namedtuple('color_config', 'range_h range_s range_v')

 
    # Original: strcut blob_data [DELETED]
    # int x    -> x
    # int y    -> y
    # int area -> area
    # BLOB_DATA = collections.namedtuple('blob_data', 'x y area')
    

    # Original: strcut blob_data_range [DELETED]
    # int x[2]    -> x
    # int y[2]    -> y
    # int area[2] -> area
    # BLOB_DATA_RANGE = collections.namedtuple('blob_data_range', 'x y area')

 
    # DESCRIPTION : the contour's area and its moments position
    #               This is used in place of struct blob_data
    # VARIABLES   : x, y, area
    CONTOUR_DATA = collections.namedtuple('contour_data', 'x y area')


    # DESCRIPTION : the range of contour's x, y and area
    # VARIABLES   : x, y, area
    CONTOUR_DATA_RANGE = collections.namedtuple('contour_data_range', 'x y area')

    def __init__(self):
        '''
        __init__(self)

        ================================================================================
        -------------------- PRIVATE VARIABLES --------------------
        int verbose_                      -> _verbose
        COLOR_CONFIG col_conf_[COLOR_NUM] -> _col_conf
        int intensity_th_                 -> _intensity_th
        int area_th_                      -> _area_th
        std::vector<BLOB_DATA> blobs_     -> _blobs       [DELETED]
        BLOB_DATA_RANGE blobs_range_      -> _blobs_range [DELETED]
        IplImage* debug_img_              -> _debug_img
        None                              -> _moments_pos [NEW VARIABLE] 
                                             [DELETED on May 2, 2014]
        None                              -> _contours
        None                              -> _contours_range

        CHANGES:
            April 29, 2014
                Delete self._blobs and self._blobs_range
            
            April 30, 2014
                Append new variable _moments_pos

            May 2, 2014
                Delete _moments_pos and append _contours and _contours_range
        ================================================================================
        '''
        self._verbose = 0
        self._col_conf = []
        self._intensity_th = 0
        self._area_th = 0
        # self._blobs = []
        # self._blobs_range = None
        self._debug_img = None


        # self._moments_pos = []
        # for i in xrange(0, self.Colors['COLOR_NUM']):
        #     self._moments_pos.append( (0.0, 0.0) )

        self._contours = []
        for i in xrange(self.Colors['COLOR_NUM']):
            self._contours.append( self.CONTOUR_DATA(0, 0, 0) )
        self._contours_range = self.CONTOUR_DATA_RANGE(x = [-1, 1],
                                                       y = [-1, 1],
                                                       area = [-1, 1])
            
    # -------------------- PUBLIC FUNCTIONS --------------------
    
    def Initialize(self, filename, verbose = 0):
        '''
        Initialize(self, filename, verbose=0) -> None

        ================================================================================
        DESCRIPTION:
            Configure the parameter. If self._verbose != 0, then show the images of
            source and destination.
        ORIGINAL:
            void Initialize(const char* filename, int verbose=0)
        CHANGES:
            const char* filename -> filename
            int verbose=0        -> verbose=0
        ================================================================================
        '''
        self._InitConfig(filename)
        self._verbose = verbose

        if self._verbose:
            cv.NamedWindow('src_image', cv.CV_WINDOW_AUTOSIZE)
            cv.MoveWindow('src_image', 50, 50)
            cv.NamedWindow('dst_image', cv.CV_WINDOW_AUTOSIZE)
            cv.MoveWindow('des_image', 50, 350)
        
    def SetVerbose(self, val):
        '''
        SetVerbose(self, val) -> None

        ================================================================================
        ORIGINAL:
            void SetVerbose(int val)
        CHANGES:
            int val -> val
        ================================================================================
        '''
        self._verbose = val

    def GetColorConfig(self, ind):
        '''
        GetColorConfig(self, ind) -> self._col_conf[ind]

        ================================================================================
        Original function:
            COLOR_CONFIG GetColorConfig(int ind) -> col_conf_[ind]
        Changes:
            int ind -> ind
        ================================================================================
        '''
        return self._col_conf[ind]

    def SetColorConfig(self, ind, conf):
        '''
        SetColorConfig(self, ind, conf) -> None

        ================================================================================
        ORIGINAL:
            void SetColorConfig(int ind, COLOR_CONFIG conf)
        CHANGES:
            int ind           -> ind
            COLOR_CONFIG conf -> conf
        ================================================================================
        '''
        self._col_conf[ind] = conf

    def Process(self, src_img):
        '''
        Process(self, src_img) -> None

        ================================================================================
        ORIGINAL:
            void Process(IplImage *src_img)
        CHANGES:
            IplImage *src_img -> src_img

            April 29, 2014
                Delete BLOBs

            May 2, 2014
                Append self._coutours and self._contours_range
        ================================================================================
        '''
        # self._blobs_range = self.BLOB_DATA_RANGE(x = [0, src_size.width],
        #                                          y = [0, src_size.height],
        #                                          area = [0, src_size.height * src_size.width / 4])
        self._contours_range = self.CONTOUR_DATA_RANGE(x    = [0, src_img.width],
                                                       y    = [0, src_img.height],
                                                       area = [0, src_img.height * src_img.width / 4])

        if self._debug_img == None:
            self._debug_img = cv.CreateImage(cv.GetSize(src_img), src_img.depth, src_img.nChannels)
        dst_img     = cv.CreateImage(cv.GetSize(src_img), src_img.depth, src_img.nChannels)
        dst_all_img = cv.CreateImage(cv.GetSize(src_img), src_img.depth, src_img.nChannels)
        cv.Zero(self._debug_img)
        cv.Zero(dst_all_img)

#        self._blobs = []
        for i in xrange(0, self.Colors['COLOR_NUM']):
            # Color extraction
            dst_img = self._ColorExtraction(src_img, cv.CV_BGR2HSV,
                                            self._col_conf[i].range_h,
                                            self._col_conf[i].range_s,
                                            self._col_conf[i].range_v)

            cv.Add(dst_all_img, dst_img, dst_all_img)

            # Labeling
            moments_x, moments_y, contour_area = self._Labeling(dst_img, self._intensity_th, i, 1)
            self._contours[i] = self.CONTOUR_DATA(x = moments_x, y = moments_y, area = contour_area)

            # Store the center of the rectangles [DELETED]
            # bd = self.BLOB_DATA(x = p1.x + (p2.x - p1.x) / 2,
            #                     y = p1.y + (p2.y - p2.y) / 2,
            #                     area = area)
            # self._blob.append(bd)
        if self._verbose:
            cv.ShowImage('src_image', src_img)
            cv.ShowImage('dst_image', dst_all_img)
            cv.NamedWindow('debug', cv.CV_WINDOW_AUTOSIZE)
            cv.MoveWindow('debug', 400, 50)
            cv.ShowImage('debug', self._debug_img)

        del dst_img
        del dst_all_img

    # def GetBlobDataSize(self):
    #     '''
    #     GetBlobDataSize(self) -> self.Colors['COLOR_NUM']
    # 
    #     ================================================================================
    #     ORIGINAL:
    #         int GetBlobDataSize() -> COLOR_NUM
    #     CHANGES:
    #         NOTHING
    #     ================================================================================
    #     '''
    #     return self.Colors['COLOR_NUM']
    # 
    # def GetBlobData(self, index):
    #     '''
    #     GetBlobData(self, index) -> self._blobs[index]
    #
    #     ================================================================================
    #     ORIGINAL:
    #         BLOB_DATA GetBlobData(const int index) -> blobs_[index]
    #     CHANGES:
    #         const int index -> index
    #     ================================================================================
    #     '''
    #     return self._blobs[index]
    #

    def GetContourDataSize(self):
        '''
        GetContourDataSize(self) -> self.Colors['COLOR_NUM']

        ================================================================================
        NEW FUNCTION ( but replacement for GetBlobDataSize() )
        ================================================================================
        '''
        return self.Colors['COLOR_NUM']
            
    def GetContourData(self, index):
        '''
        GetContourData(self, index) -> self._contours[index]

        ================================================================================
        NEW FUNCTION ( replacement for GetBlobData(const int index) )
        ================================================================================
        '''
        return self._contours[index]

    def GetXMin(self):
        '''
        GetXMin(self) -> self._contours_range.x[0]
    
        ================================================================================
        ORIGINAL:
            int GetXMin() -> blobs_range_.x[0]
        CHANGES:
            Use contour replacement for blob
        ================================================================================
        '''
        return self._contours_range.x[0]

    def GetXMax(self):
        '''
        GetXMax(self) -> self._contours_range.x[1]
    
        ================================================================================
        ORIGINAL:
            int GetXMax() -> blobs_range_.x[1]
        CHANGES:
            Use contour replacement for blob
        ================================================================================
        '''
        return self._contours_range.x[1]
    
    def GetYMin(self):
        '''
        GetYMin(self) -> self._contours_range.y[0]
   
        ================================================================================
        ORIGINAL:
            int GetYMin() -> blobs_range_.y[0]
        CHANGES:
            Use contour replacement for blob
        ================================================================================
        '''
        return self._contours_range.y[0]

    def GetYMax(self):
        '''
        GetYMax(self) -> self._contours_range.y[1]
   
        ================================================================================
        ORIGINAL:
            int GetYMax() -> blobs_range_.y[1]
        CHANGES:
            Use contour replacement for blob
        ================================================================================
        '''
        return self._contours_range.y[1]

    def GetAreaMin(self):
        '''
        GetAreaMin(self) -> self._contours_range.area[0]
   
        ================================================================================
        ORIGINAL:
            int GetAreaMin() -> blobs_range_.area[0]
        CHANGES:
            Use contour replacement for blob
        ================================================================================
        '''
        return self._contours_range.area[0]

    def GetAreaMax(self):
        '''
        GetAreaMax(self) -> self._contours_range.area[1]
   
        ================================================================================
        ORIGINAL:
            int GetAreaMax() -> blobs_range_.area[1]
        CHANGES:
            Use contour replacement for blob
        ================================================================================
        '''
        return self._contours_range.area[1]

    def GetDebugImg(self):
        '''
        GetDebugImg(self) -> self._debug_img

        ================================================================================
        ORIGINAL:
            IplImage* GetDebugImg() -> debug_img_
        CHANGES:
            NOTHING
        ================================================================================
        '''
        return self._debug_img


    #def GetMomentsPos(self, ind):
    #    '''
    #    GetMomentsPos(self, ind) -> self._moments_pos[ind]
    #
    #    ================================================================================
    #    This is NEW FUNCTION. This returns the position of moments of the color region.
    #    ================================================================================
    #    '''
    #    return self._moments_pos[ind]

    def CalcHist(self, src_img):
        '''
        CalcHist(self, src_img) -> None

        ================================================================================
        ORIGINAL:
            void CalcHist(IplImage *src_img, int *range_h, int *range_s)
        CHANGES:
            IplImage *src_img -> src_img
            int *range_h      -> None
            int *range_s      -> None
            
            April 24, 2014 -> [CAUSION] usage of cv.CreateHist
                in the online document
                C or C++: cvHistgram* cvCreateHist(int dims, int* sizes, int type,
                                                   float** ranges=NULL, int uniform=1)
                Python  : cv.CreateHist(dims, type, ranges=None, uniform=1) -> hist
                It says in the online document or helps that the first argument of 
                cv.CreateHist is 'dims', but actually is 'sizes.'
                (referenced to http://www.nullege.com/codes/search/cv.CreateHist
                 cited by April 24, 2014)
                So in Python, usage of the cv.CreateHist is
                    cv.CreateHist(sizes, type, ranges=None, uniform=1)

            April 30, 2014
                The private method '_Labeling' was changed and it returns only 'area.'
                Therefore, printing and computing heu and saturation is deleted.
        ================================================================================
        '''
        h_plane = cv.CreateImage(cv.GetSize(src_img), 8, 1)
        s_plane = cv.CreateImage(cv.GetSize(src_img), 8, 1)
        v_plane = cv.CreateImage(cv.GetSize(src_img), 8, 1)
        planes = [h_plane, s_plane]
        hsv = cv.CreateImage(cv.GetSize(src_img), 8, 3)
        h_bins = 180 # int
        s_bins = 255 # int
        hist_size = [h_bins, s_bins]
        h_ranges = [0.0, 180.0] # float
        s_ranges = [0.0, 255.0] # float
        ranges = [h_ranges, s_ranges]
        scale = 2 # int
        hist_img = cv.CreateImage((h_bins * scale, s_bins * scale), 8, 3)
        # CvHistgram hist
        # float max_value = 0.0
        # int h, s

        cv.CvtColor(src_img, hsv, cv.CV_BGR2HSV)
        cv.Split(hsv, h_plane, s_plane, v_plane, None)
        
        # hist = cvCreateHist(2, hist_size, cv.CV_HIST_ARRAY, ranges, 1)
        hist = cv.CreateHist(hist_size, cv.CV_HIST_ARRAY, ranges, 1)
        cv.CalcHist(planes, hist, 0, None)
        
        # cv.GetMinMaxHistValue(hist) -> (min_value, max_value, min_idx, max_idx)
        # only use 'max_value'
        min_value, max_value, min_idx, max_idx = cv.GetMinMaxHistValue(hist)
        cv.Zero(hist_img)
        
        for h in xrange(0, h_bins):
            for s in xrange(0, s_bins):
                bin_val = float( cv.GetReal2D(hist.bins, h, s) )
                intensity = cv.Round(bin_val * 255 / max_value)
                cv.Rectangle(hist_img, (h*scale, s*scale),
                             ( (h+1)*scale - 1, (s+1)*scale - 1),
                             cv.CV_RGB(intensity, intensity, intensity),
                             cv.CV_FILLED)

        cv.NamedWindow('H-S Histgram', cv.CV_WINDOW_AUTOSIZE)
        cv.MoveWindow('H-S Histgram', 400, 350)
        cv.ShowImage('H-S Histgram', hist_img)


        # area = self._Labeling(hist_img, self._intensity_th)
        # print 'H:%d-%d S:%d-%d\n' % (p1.x/scale, p2.x/scale, p1.y/scale, p2.y/scale)
        
        # range_h = [p1.x/2, p2.x/2]
        # range_s = [p1.y/2, p2.y/2]

        # cv.DestroyWindow('H-S Histgram')

        del hist
        del hist_img
        del hsv
        del h_plane
        del s_plane
        del v_plane
        # return range_h, range_s

    def SaveParameters(self, filename):
        '''
        SaveParameters(self, filename)

        ================================================================================
        DESCRIPTION:
            Save parametes of colors, threshold of intensity and threshold of area
        ORIGINAL:
            int SaveParameters(const char *filename)
        CHANGES:
            const char *filename -> filename
        ================================================================================
        '''
        try:
            fpw = open(filename, 'w')
        except:
            print 'Config file open error'
            sys.exit(1)

        for i in xrange(0, self.Colors['COLOR_NUM']):
            fpw.write('%s %d %d %d %d %d %d\n' %
                      (self.color_name[i],
                       self._col_conf[i].range_h[0], self._col_conf[i].range_h[1],
                       self._col_conf[i].range_s[0], self._col_conf[i].range_s[1],
                       self._col_conf[i].range_v[0], self._col_conf[i].range_v[1]))
        fpw.write('INTNS_TH %d\n' % self._intensity_th)
        fpw.write('AREA_TH %d\n' % self._area_th)

        fpw.close()

    # -------------------- PRIVATE FUNCTIONS --------------------

    def _InitConfig(self, filename):
        '''
        _InitConfig(self, filename) -> None

        ================================================================================
        ORIGINAL:
            void InitConfig(const char* filename)
        CHANGES:
            const char* filename -> filename
        ================================================================================
        '''
        for i in range(0, self.Colors['COLOR_NUM']):
            self._col_conf.append(self.COLOR_CONFIG(range_h = [0, 0],
                                                    range_s = [0, 0],
                                                    range_v = [0, 0]))
        self._LoadParameters(filename)
        
    def _LoadParameters(self, filename):
        '''
        _LoadParameters(self, filename) -> 0
        
        ================================================================================
        ORIGINAL:
            int LoadParameters(const char *filename) -> 0
        CHANGES:
            const char *filename -> filename
        ================================================================================
        '''
        try:
            fpr = open(filename, 'r')
        except:
            print 'Config file open error!!'
            sys.exit(1)
        while True:
            string = fpr.readline()
            if string == '':
                break
            if not string[0] == '\n':
                arg = re.split('[ ]+', string)
                if not arg[0][0] == '#':
                    self._SetParameter(arg)
        fpr.close()
        return 0

    def _SetParameter(self, arg):
        '''
        _SetParameter(self, arg) -> 0

        ================================================================================
        DESCRIPTION:
            Set parameters of color, threshold of intensity and threshold of area.
        ORIGINAL:
            int SetParameter(char **arg) -> 0
        CHANGES:
            char **arg -> arg
        ================================================================================
        '''
        if arg[0] in self.color_name:
            i = self.Colors[arg[0]]
        elif arg[0] == 'INTNS_TH':
            self._intensity_th = int(arg[1])
            return 0
        elif arg[0] == 'AREA_TH':
            self._area_th = int(arg[1])
            return 0
        else:
            print 'unknown parameter tag %s\n' % (arg[0])
            return 0
        self._col_conf[i] = self.COLOR_CONFIG(range_h = [int(arg[1]), int(arg[2])],
                                              range_s = [int(arg[3]), int(arg[4])],
                                              range_v = [int(arg[5]), int(arg[6])])
        return 0

    def _ColorExtraction(self, src_img, code, range_h, range_s, range_v): # -> dst_img
        '''
        _ColorExtraction(self, src_img, code, range_h, range_s, range_v) -> dst_img

        ================================================================================
        DESCRIPTION:
            Extract the color within range_h, range_s and range_v.
            The third argument 'code' is color space conversion code of OpenCV
        ORIGINAL:
            void ColorExtraction(IplImage* src_img, IplImage* dst_img,
                                 int code,
                                 int *range_h, int *range_s, int *range_v)
        CHANGES:
            IplImage* src_img -> src_img
            IplImage* dst_img -> None(return value)
            int code          -> code
            int *range_h      -> range_h
            int *range_s      -> range_s
            int *range_v      -> range_v
        ================================================================================
        '''
        lower = np.zeros(3)
        upper = np.zeros(3)
        val   = np.zeros(3)

        # Color format conversion
        Color_img = cv.CreateImage(cv.GetSize(src_img), src_img.depth, src_img.nChannels)
        cv.CvtColor(src_img, Color_img, code)

        # Create 3ch LUT (Loock Up Table)
        lut = cv.CreateMat(256, 1, cv.CV_8UC3)

        lower[:] = [range_h[0], range_s[0], range_v[0]]
        upper[:] = [range_h[1], range_s[1], range_v[1]]

        for i in xrange(0, 256):
            for k in xrange(0, 3):
                if lower[k] <= upper[k]:
                    if (lower[k] <= i) and (i <= upper[k]):
                        val[k] = 255
                    else:
                        val[k] = 0
                else:
                    if (i <= upper[k]) or (lower[k] <= i):
                        val[k] = 255
                    else:
                        val[k] = 0
            # Set LUT
            cv.Set1D(lut, i, cv.Scalar(val[0], val[1], val[2]))

        cv.LUT(Color_img, Color_img, lut)
        # python-opencv don't have cvReleaseMat(mat)
        # cv.ReleaseMat(lut)
        del lut

        # Allocate IplImage memory for the 3 channels
        ch1_img = cv.CreateImage(cv.GetSize(Color_img), Color_img.depth, 1)
        ch2_img = cv.CreateImage(cv.GetSize(Color_img), Color_img.depth, 1)
        ch3_img = cv.CreateImage(cv.GetSize(Color_img), Color_img.depth, 1)

        # Split image
        cv.Split(Color_img, ch1_img, ch2_img, ch3_img, None)
        
        # Take AND of all channels and make a mask image
        Mask_img = cv.CreateImage(cv.GetSize(Color_img), Color_img.depth, 1)
        cv.And(ch1_img, ch2_img, Mask_img)
        cv.And(Mask_img, ch3_img, Mask_img)

        # Copy src_img to dst_img using the mask image
        dst_img = cv.CreateImage(cv.GetSize(src_img), src_img.depth, src_img.nChannels)
        cv.Zero(dst_img)
        cv.Copy(src_img, dst_img, Mask_img)

        del Color_img
        del ch1_img
        del ch2_img
        del ch3_img
        del Mask_img
        return dst_img

    def _Labeling(self, src_img, threshold,
                 c_ind = 0, debug_view = 0): # -> moments_x, moments_y, area_max
        '''
        _Labeling(self, src_img, threshold, c_ind, debug_view) -> moments_x, moments_y, area_max 

        ================================================================================
        DESCRIPTION:
            Label the color regions and return the maximum area of those regions.
            The fourth argument 'c_ind' is index of colors. If 'debug_view' == 1,
            then draw the contours and the moment of the region on self._debug_img.
        ORIGINAL:
            void Labeling(IplImage *src_img,
                          int param1, int param2,
                          CvPoint& p1_max, CvPoint& p2_max, int& area,
                          int c_ind = 0, int debug_view = 0)
        CHANGES:
            IplImage *src_img -> src_img
            int param1        -> threshold
            int param2        -> NOTHING
            CvPoint& p1_max   -> NOTHING
            CvPoint& p2_max   -> NOTHING
            int& area         -> NOTHING (return value as area_max)
            int c_ind         -> c_ind (optional default = 0)
            int debug_view    -> debug_view (optional default = 0)
        ================================================================================
        '''
        work_img = cv.CreateImage(cv.GetSize(src_img), cv.IPL_DEPTH_8U, 1)
        cv.CvtColor(src_img, work_img, cv.CV_BGR2GRAY)
        cv.Threshold(work_img, work_img, threshold, 255, cv.CV_THRESH_BINARY)

        storage = cv.CreateMemStorage()
        contours = cv.FindContours(work_img, storage, cv.CV_RETR_LIST, cv.CV_CHAIN_APPROX_NONE)

        del work_img
        del storage

        area_max = 0
        area_max_moments = None
        area_max_index = 0
        index = 0
        work_contours = contours
        while work_contours != None:
            contour = work_contours[:]
            if contour == []:
                break
            area = cv.ContourArea(work_contours)
            if area_max < area:
                area_max = area
                area_max_moments = cv.Moments(work_contours)
                area_max_index = index
            if work_contours.h_next() is None:
                break
            work_contours = work_contours.h_next()
            index += 1
        for i in xrange(0, area_max_index):
            contours = contours.h_next()

        del work_contours

        moments_x = 0
        moments_y = 0
        if debug_view == 1:
            if not contours is None:
                cvColors = [cv.CV_RGB(255, 0, 0),   # RED
                            cv.CV_RGB(255, 255, 0), # YELLOW
                            cv.CV_RGB(0, 255, 0),   # GREEN
                            cv.CV_RGB(0, 0, 255)]   # BLUE
                if self.color_name[c_ind] in self.Colors:
                    cv.DrawContours(self._debug_img, contours,
                                    cvColors[self.Colors[self.color_name[c_ind]]],
                                    cvColors[self.Colors[self.color_name[c_ind]]], 0)
                    if not area_max_moments is None:
                        moments_x = int(area_max_moments.m10 / area_max_moments.m00)
                        moments_y = int(area_max_moments.m01 / area_max_moments.m00)
                        cv.Circle(self._debug_img, (moments_x, moments_y),
                                  10, # radius
                                  cvColors[self.Colors[self.color_name[c_ind]]], 
                                  -1) # thickness / thickness = -1 means the filled circle is given

        del contours
        del area_max_moments

        return moments_x, moments_y, area_max
